"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const tslib_1 = require("tslib");
const array_1 = require("../utils/array");
const format_1 = require("../utils/format");
const colors_1 = require("./colors");
const validators_1 = require("./validators");
const DEFAULT_DOTS_WIDTH = 25;
class HelpFormatter {
    constructor({ colors }) {
        this.colors = colors ? colors : colors_1.DEFAULT_COLORS;
    }
}
exports.HelpFormatter = HelpFormatter;
class NamespaceHelpFormatter extends HelpFormatter {
    constructor({ location, namespace, colors }) {
        super({ colors });
        this.dotswidth = DEFAULT_DOTS_WIDTH;
        this.location = location;
        this.namespace = namespace;
    }
    getNamespaceMetadata() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._metadata) {
                this._metadata = yield this.namespace.getMetadata();
            }
            return this._metadata;
        });
    }
    getNamespaceFullName() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._fullName) {
                this._fullName = this.location.path.map(([p]) => p).join(' ');
            }
            return this._fullName;
        });
    }
    formatHeader() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong, input } = this.colors;
            const fullName = yield this.getNamespaceFullName();
            const description = yield this.formatDescription();
            const longDescription = yield this.formatLongDescription();
            return (`\n  ${strong(`${input(fullName)}${description}`)}` +
                (longDescription ? `\n\n    ${longDescription}` : '') + '\n');
        });
    }
    formatDescription() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const fullName = yield this.getNamespaceFullName();
            const metadata = yield this.getNamespaceMetadata();
            const wrappedDescription = format_1.wordWrap(metadata.description, { indentation: fullName.length + 5 });
            const subtitle = yield this.formatSubtitle();
            const description = (subtitle ? subtitle + ' ' : '') + wrappedDescription;
            return description ? ` - ${description}` : '';
        });
    }
    formatSubtitle() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return '';
        });
    }
    formatLongDescription() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const metadata = yield this.getNamespaceMetadata();
            if (!metadata.longDescription) {
                return '';
            }
            return format_1.wordWrap(metadata.longDescription.trim(), { indentation: 4 });
        });
    }
    getExtraOptions() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return [];
        });
    }
    formatUsage() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong, weak, input } = this.colors;
            const fullName = yield this.getNamespaceFullName();
            const options = ['--help', ...(yield this.getExtraOptions())];
            const usageLines = [
                `<command> ${weak('[<args>]')} ${options.map(opt => weak('[' + opt + ']')).join(' ')} ${weak('[options]')}`,
            ];
            return (`\n  ${strong('Usage')}:` +
                `\n\n    ${usageLines.map(u => `${weak('$')} ${input(fullName + ' ' + u)}`).join('\n    ')}\n`);
        });
    }
    formatCommands() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const commands = yield this.namespace.getCommandMetadataList();
            return this.formatCommandGroup('Commands', commands);
        });
    }
    formatCommandGroup(title, commands) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong } = this.colors;
            const filterCallback = this.filterCommandCallback;
            const filteredCommands = filterCallback ? yield array_1.filter(commands, (cmd) => tslib_1.__awaiter(this, void 0, void 0, function* () { return filterCallback(cmd); })) : commands;
            const [cmdDetails, nsDetails] = yield Promise.all([
                this.getListOfCommandDetails(filteredCommands.filter(cmd => cmd.namespace === this.namespace)),
                this.getListOfNamespaceDetails(filteredCommands.filter(cmd => cmd.namespace !== this.namespace)),
            ]);
            const details = [...cmdDetails, ...nsDetails];
            if (details.length === 0) {
                return '';
            }
            details.sort();
            return (`\n  ${strong(title)}:` +
                `\n\n    ${details.join('\n    ')}\n`);
        });
    }
    getListOfCommandDetails(commands) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { weak, input } = this.colors;
            const fullCmd = commands.map(cmd => cmd.path.map(([p]) => p).join(' '));
            const fillStringArray = format_1.generateFillSpaceStringList(fullCmd, this.dotswidth, weak('.'));
            const formattedCommands = yield Promise.all(commands.map((cmd, index) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                const description = (yield this.formatBeforeCommandDescription(cmd)) + cmd.description + (yield this.formatAfterCommandDescription(cmd));
                const wrappedDescription = format_1.wordWrap(description, { indentation: this.dotswidth + 6 });
                return `${input(cmd.path.map(p => p[0]).join(' '))}${wrappedDescription ? ' ' + fillStringArray[index] + ' ' + wrappedDescription : ''}`;
            })));
            return formattedCommands;
        });
    }
    getListOfNamespaceDetails(commands) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { weak, input } = this.colors;
            const descriptions = new Map();
            const grouped = new Map();
            yield Promise.all(commands.map((cmd) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                const nsmeta = yield cmd.namespace.getMetadata();
                descriptions.set(nsmeta.name, nsmeta.description);
                let entry = grouped.get(nsmeta.name);
                if (!entry) {
                    entry = { meta: nsmeta, cmds: [] };
                    grouped.set(nsmeta.name, entry);
                }
                entry.cmds.push(cmd);
            })));
            const entries = [...grouped.entries()];
            const fillStringArray = format_1.generateFillSpaceStringList(entries.map(([name]) => name + ' <subcommand>'), this.dotswidth, weak('.'));
            const formattedNamespaces = yield Promise.all(entries.map(([name, { meta, cmds }], i) => tslib_1.__awaiter(this, void 0, void 0, function* () {
                const subcommands = cmds.map(c => input(c.name)).join(', ');
                const wrappedDescription = format_1.wordWrap(`${yield this.formatBeforeNamespaceDescription(meta)}${descriptions.get(name)} ${weak('(subcommands:')} ${subcommands}${weak(')')}`, { indentation: this.dotswidth + 6 });
                return `${input(name + ' <subcommand>')}${wrappedDescription ? ' ' + fillStringArray[i] + ' ' + wrappedDescription : ''}`;
            })));
            return formattedNamespaces;
        });
    }
    formatBeforeNamespaceDescription(meta) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return '';
        });
    }
    formatBeforeCommandDescription(cmd) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return '';
        });
    }
    formatAfterCommandDescription(cmd) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { weak, input } = this.colors;
            return `${cmd.aliases.length > 0 ? weak(' (alias' + (cmd.aliases.length === 1 ? '' : 'es') + ': ') + cmd.aliases.map(a => input(a)).join(', ') + weak(')') : ''}`;
        });
    }
    format() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return ((yield this.formatHeader()) +
                (yield this.formatUsage()) +
                (yield this.formatCommands()) +
                '\n');
        });
    }
}
exports.NamespaceHelpFormatter = NamespaceHelpFormatter;
class CommandHelpFormatter extends HelpFormatter {
    constructor({ location, command, colors }) {
        super({ colors });
        this.dotswidth = DEFAULT_DOTS_WIDTH;
        this.location = location;
        this.command = command;
    }
    getCommandMetadata() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._metadata) {
                this._metadata = yield this.command.getMetadata({ location: this.location });
            }
            return this._metadata;
        });
    }
    getCommandFullName() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (!this._fullName) {
                this._fullName = this.location.path.map(([p]) => p).join(' ');
            }
            return this._fullName;
        });
    }
    filterOptionCallback(options) {
        return true;
    }
    formatHeader() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong, input } = this.colors;
            const fullName = yield this.getCommandFullName();
            const description = yield this.formatDescription();
            const longDescription = yield this.formatLongDescription();
            return (`\n  ${strong(`${input(fullName)}${description}`)}` +
                (longDescription ? `\n\n    ${longDescription}` : '') + '\n');
        });
    }
    formatDescription() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const fullName = yield this.getCommandFullName();
            const metadata = yield this.getCommandMetadata();
            const wrappedDescription = format_1.wordWrap(metadata.description, { indentation: fullName.length + 5 });
            const subtitle = yield this.formatSubtitle();
            const description = (subtitle ? subtitle + ' ' : '') + wrappedDescription;
            return description ? ` - ${description}` : '';
        });
    }
    formatSubtitle() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return '';
        });
    }
    formatLongDescription() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const metadata = yield this.getCommandMetadata();
            if (!metadata.longDescription) {
                return '';
            }
            return format_1.wordWrap(metadata.longDescription.trim(), { indentation: 4 });
        });
    }
    formatInlineInput(input) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            if (input.validators && input.validators.includes(validators_1.validators.required)) {
                return '<' + input.name + '>';
            }
            return '[<' + input.name + '>]';
        });
    }
    formatUsage() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong, weak, input } = this.colors;
            const fullName = yield this.getCommandFullName();
            const metadata = yield this.getCommandMetadata();
            const options = metadata.options ? metadata.options.filter(opt => this.filterOptionCallback(opt)) : [];
            const formattedInputs = metadata.inputs ? yield Promise.all(metadata.inputs.map((input) => tslib_1.__awaiter(this, void 0, void 0, function* () { return this.formatInlineInput(input); }))) : [];
            return (`\n  ${strong('Usage')}:` +
                `\n\n    ${weak('$')} ${input(fullName + (formattedInputs ? ' ' + formattedInputs.join(' ') : ''))}${options.length > 0 ? ' ' + input('[options]') : ''}\n`);
        });
    }
    formatInputs() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong, weak, input } = this.colors;
            const metadata = yield this.getCommandMetadata();
            const inputs = metadata.inputs ? metadata.inputs : [];
            if (inputs.length === 0) {
                return '';
            }
            const fillStrings = format_1.generateFillSpaceStringList(inputs.map(input => input.name), this.dotswidth, weak('.'));
            const inputLineFn = ({ name, description }, index) => {
                const optionList = input(`${name}`);
                const wrappedDescription = format_1.wordWrap(description, { indentation: this.dotswidth + 6 });
                return `${optionList} ${fillStrings[index]} ${wrappedDescription}`;
            };
            return (`\n  ${strong('Inputs')}:` +
                `\n\n    ${inputs.map(inputLineFn).join('\n    ')}\n`);
        });
    }
    formatOptionLine(opt) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { weak, input } = this.colors;
            const showInverse = opt.type === Boolean && opt.default === true && opt.name.length > 1;
            const optionList = (showInverse ? input(`--no-${opt.name}`) : input(`-${opt.name.length > 1 ? '-' : ''}${opt.name}`)) +
                (!showInverse && opt.aliases && opt.aliases.length > 0 ? ', ' +
                    opt.aliases
                        .map(alias => input(`-${alias}`))
                        .join(', ') : '');
            const optionListLength = format_1.stringWidth(optionList);
            const fullLength = optionListLength > this.dotswidth ? optionListLength + 1 : this.dotswidth;
            const fullDescription = ((yield this.formatBeforeOptionDescription(opt)) +
                opt.description +
                (yield this.formatAfterOptionDescription(opt)));
            const wrappedDescription = format_1.wordWrap(fullDescription, { indentation: this.dotswidth + 6 });
            return `${optionList} ${weak('.').repeat(fullLength - optionListLength)} ${wrappedDescription}`;
        });
    }
    formatBeforeOptionDescription(opt) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return '';
        });
    }
    formatAfterOptionDescription(opt) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return this.formatOptionDefault(opt);
        });
    }
    formatOptionDefault(opt) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { weak, input } = this.colors;
            if (typeof opt.default === 'string') {
                return weak(' (default: ') + input(opt.default) + weak(')');
            }
            else {
                return '';
            }
        });
    }
    formatOptions() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const metadata = yield this.getCommandMetadata();
            const options = metadata.options ? metadata.options : [];
            return this.formatOptionsGroup('Options', options);
        });
    }
    formatOptionsGroup(title, options) {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong } = this.colors;
            options = options.filter(opt => this.filterOptionCallback(opt));
            if (options.length === 0) {
                return '';
            }
            const formattedOptions = yield Promise.all(options.map((option) => tslib_1.__awaiter(this, void 0, void 0, function* () { return this.formatOptionLine(option); })));
            return (`\n  ${strong(title)}:` +
                `\n\n    ${formattedOptions.join('\n    ')}\n`);
        });
    }
    formatExamples() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            const { strong, weak, input } = this.colors;
            const metadata = yield this.getCommandMetadata();
            const fullName = yield this.getCommandFullName();
            if (!metadata.exampleCommands || !Array.isArray(metadata.exampleCommands)) {
                return '';
            }
            const exampleLines = metadata.exampleCommands.map(cmd => {
                const sepIndex = cmd.indexOf(' -- ');
                cmd = sepIndex === -1 ? input(cmd) : input(cmd.substring(0, sepIndex)) + cmd.substring(sepIndex);
                const wrappedCmd = format_1.wordWrap(cmd, { indentation: 12, append: ' \\' });
                return `${weak('$')} ${input(fullName)}${wrappedCmd ? ' ' + wrappedCmd : ''}`;
            });
            return (`\n  ${strong('Examples')}:` +
                `\n\n    ${exampleLines.join('\n    ')}\n`);
        });
    }
    format() {
        return tslib_1.__awaiter(this, void 0, void 0, function* () {
            return ((yield this.formatHeader()) +
                (yield this.formatUsage()) +
                (yield this.formatInputs()) +
                (yield this.formatOptions()) +
                (yield this.formatExamples()) +
                '\n');
        });
    }
}
exports.CommandHelpFormatter = CommandHelpFormatter;
